# -*- coding: utf-8 -*-

"""Top-level package for Neat Panda."""

__author__ = """Henric Sundberg"""
__email__ = "henric.sundberg@gmail.com"
__version__ = "0.7.1"

from ._tidy import spread, gather
from ._caretaker import clean_column_names, _clean_column_names
