#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['neat_panda']

package_data = \
{'': ['*']}

install_requires = \
['pandas', 'pandas-flavor']

extras_require = \
{'test': ['pytest', 'pytest-runner']}

setup(name='neat_panda',
      version='0.7.1',
      description='Top-level package for Neat Panda.',
      author='Henric Sundberg',
      author_email='henric.sundberg@gmail.com',
      url='https://github.com/htp84/neat_panda',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
